import re
import logging
from pathlib import Path
from typing import Optional, Tuple

from toml import load, dump
from openapi_client import Configuration, ApiClient
from openapi_client.models import KeyLoginParams, UserData
from openapi_client.apis import DefaultApi
from leapcli.exceptions import MalformedKeys, KeysMixedUp, LoginFailed, CredentialsNotFound
from leapcli.project import Project

CREDENTIALS_FILENAME = 'credentials.toml'

_log = logging.getLogger(__name__)


class _Authenticator:
    def __init__(self):
        self._api_id: Optional[str] = None
        self._api_key: Optional[str] = None
        self._api_client: Optional[DefaultApi] = None
        self._project: Optional[Project] = None
        self.user: Optional[UserData] = None
        self.cookie: Optional[str] = None
        self.is_initialized = False

    def initialize(self, project: Project, api_id: Optional[str] = None,
                   api_key: Optional[str] = None, should_write_credentials=False):
        self._project = project
        self._api_key = api_key
        self._api_id = api_id
        if not api_key or not api_id:
            self._api_id, self._api_key = _Authenticator.read_credentials()

        if not self._api_key or not self._api_id:
            raise CredentialsNotFound()

        if not _Authenticator.is_valid_api_id(self._api_id):
            if _Authenticator.is_valid_api_key(self._api_id):
                raise KeysMixedUp()
            raise MalformedKeys()

        if not _Authenticator.is_valid_api_key(self._api_key):
            raise MalformedKeys()

        self.authenticated_api()

        if should_write_credentials:
            self.write_credentials(self._api_id, self._api_key)

        self.is_initialized = True

    @staticmethod
    def redact_key(key: str) -> str:
        assert len(key) > 3
        # Mask the first 19 characters of a key
        return '*' * 19 + key[-3:]

    @staticmethod
    def is_valid_api_key(api_key: str) -> bool:
        return re.match(r'^k0\w{20}$', api_key) is not None

    @staticmethod
    def is_valid_api_id(api_id: str) -> bool:
        return re.match(r'^i0\w{20}$', api_id) is not None

    @staticmethod
    def credentials_file_path() -> Path:
        return Project.config_dir().joinpath(CREDENTIALS_FILENAME)

    @staticmethod
    def has_credentials() -> bool:
        api_id, api_key = _Authenticator.read_credentials()
        return api_id is not None and api_key is not None

    @staticmethod
    def read_credentials() -> Tuple[Optional[str], Optional[str]]:
        # TODO: more robust handling of corrupted file
        path = _Authenticator.credentials_file_path()
        if path.is_file():
            _log.debug('reading credentials from %s', path)
            with path.open('r') as f:
                dictionary = load(f)
                return dictionary['api_id'], dictionary['api_key']
        return None, None

    @staticmethod
    def write_credentials(api_id: str, api_key: str) -> None:
        _log.info('writing credentials')
        with _Authenticator.credentials_file_path().open('w') as f:
            return dump(dict(api_id=api_id, api_key=api_key), f)

    def key_login(self) -> None:
        host = self._project.detect_backend_url()
        cfg = Configuration(host=host)
        unauthenticated_client = ApiClient(cfg)
        api = DefaultApi(unauthenticated_client)
        params = KeyLoginParams(self._api_id, self._api_key)
        user, status, headers = api.key_login(params, _return_http_data_only=False)
        if status != 200 or not 'Set-Cookie' in headers:
            _log.info('login failed with api_id: %s, api_key: %s', self._api_id,
                      self.redact_key(self._api_key))
            raise LoginFailed()
        self.user = user
        self.cookie = headers['Set-Cookie']

    def logged_in(self) -> bool:
        return self.cookie is not None

    def authenticated_api(self) -> DefaultApi:
        if not self.logged_in():
            self.key_login()
        host = self._project.detect_backend_url()
        cfg = Configuration(host=host)
        cookie_client = ApiClient(cfg, cookie=self.cookie)

        return DefaultApi(cookie_client)


Authenticator = _Authenticator()



import enum
import os
import logging
import tempfile
from pathlib import Path
import re
from shutil import copytree
from typing import Optional

from toml import load, dump
from git import Repo
from git.exc import InvalidGitRepositoryError
import toml
from openapi_client.apis import DefaultApi

from leapcli.exceptions import InvalidOrgName, CredentialsNotFound, AlreadyInitialized, InvalidProjectName


class Framework(str, enum.Enum):
    TENSORFLOW = 'tensorflow'
    PYTORCH = 'pytorch'


TENSORLEAP_DIR = '.tensorleap'
CONFIG_DIR = '~/.config/tensorleap'

VALID_PROJECT_REGEX = r'^[a-zA-Z0-9_\-. ]{3,}$'
VALID_PROJECT_EXPL = '''* At least 3 characters long.
* Allowed characters: alphanumeric, "_", "-", "."'''

VALID_ORG_REGEX = r'^[a-zA-Z0-9][a-zA-Z0-9\-]+[a-zA-Z0-9]$'
VALID_ORG_EXPL = '''* At least 3 characters long.
* Allowed characters: alphanumeric, "-"
* Does not start or end with a hyphen.
* No double hypens.'''

CONFIG_FILENAME = 'config.toml'

TENSORLEAP_BACKEND_DOMAIN = 'tensorleap.ai'

API_ENDPOINT = 'api/v2'

_log = logging.getLogger(__name__)


class Project:
    def __init__(self, directory: str = '.'):
        self.directory: Path = Path(directory)
        self.project: Optional[str] = None
        self.org: Optional[str] = None
        self.framework: Framework = Framework.TENSORFLOW
        self.state = {}
        self.dataset: Optional[str] = None
        self.bucket: Optional[str] = None
        self.override_url: Optional[str] = None

    def detect_project_dir(self) -> Path:
        try:
            repo = Repo(self.directory, search_parent_directories=True)
            return repo.working_tree_dir
        except InvalidGitRepositoryError:
            return os.getcwd()

    def detect_project(self) -> str:
        if self.is_initialized():
            return self.project_config()['projectName']
        raise CredentialsNotFound()

    def detect_dataset(self) -> str:
        if self.is_initialized():
            return self.project_config()['datasetName']
        raise CredentialsNotFound()

    def detect_bucket(self) -> str:
        if self.is_initialized():
            return self.project_config()['bucketName']
        raise CredentialsNotFound()

    @staticmethod
    def prompt_dataset() -> str:
        return input('Dataset name: ')

    @staticmethod
    def prompt_project() -> str:
        return input('Project name: ')

    def org_domain(self) -> str:
        return f'{self.detect_org()}.{TENSORLEAP_BACKEND_DOMAIN}'

    def detect_backend_url(self) -> str:
        assert self.is_initialized()
        if 'apiEndpoint' in self.project_config() and self.project_config()['apiEndpoint']:
            _log.debug('found apiEndpoint override in config')
            return self.project_config()['apiEndpoint']
        return f'{self.api_url()}/api/v2'

    def project_id(self, api: DefaultApi) -> str:
        # TODO: cache this result
        matches = [proj.id for proj in api.get_projects().data if proj.name == self.detect_project()]
        assert len(matches) == 1
        return matches[0]

    def dataset_id(self, api: DefaultApi) -> str:
        # TODO: cache this result
        matches = [dataset.id for dataset in api.get_datasets().datasets if dataset.name == self.detect_dataset()]
        assert len(matches) == 1
        return matches[0]

    def api_url(self) -> str:
        return f'https://api.{self.org_domain()}'

    def cache_dir(self) -> Path:
        self.read_state()
        if 'cache_dir' in self.state:
            return Path(self.state['cache_dir'])
        cache_dir = tempfile.mkdtemp(prefix='tensorleap-cache')
        self.state['cache_dir'] = cache_dir
        self.save_state()
        return Path(cache_dir)

    @staticmethod
    def leapcli_state_file() -> Path:
        return Project.config_dir().joinpath('state.toml')

    def read_state(self) -> dict:
        if Project.leapcli_state_file().is_file():
            with open(Project.leapcli_state_file(), encoding='utf-8') as f:
                self.state = load(f)
        return self.state

    def save_state(self) -> None:
        with open(Project.leapcli_state_file(), 'w', encoding='utf-8') as f:
            dump(self.state, f)

    @staticmethod
    def leapcli_package_dir() -> Path:
        return Path(__file__).parent.parent

    @staticmethod
    def leapcli_package_info() -> dict:
        poetry_conf_file = Project.leapcli_package_dir(). \
            joinpath('pyproject.toml')
        with open(poetry_conf_file, encoding='utf-8') as f:
            return toml.load(f)

    @staticmethod
    def template_dir() -> Path:
        return Project.leapcli_package_dir().joinpath('templates')

    @staticmethod
    def org_from_remote_url(url: str) -> str:
        if url is None:
            return None

        # Expect pattern: git@some.domain:tensorleap/cli.git
        git_url_pattern = r'git@(?:\w+\.)+\w+:([^/]+)/.*'
        git_url_match = re.match(git_url_pattern, url)
        if git_url_match is not None:
            return git_url_match[1]

        # Expect https://gitlab.com/gitlab-org/gitlab-foss.git
        https_url_pattern = r'https://(?:\w+\.)+\w+/([^/]+)/.*'
        https_url_match = re.match(https_url_pattern, url)
        if https_url_match:
            return https_url_match[1]
        return None

    def detect_org(self) -> str:
        if self.is_initialized():
            return self.project_config()['organization']
        try:
            repo = Repo(self.directory)
            origin_remote = next((r for r in repo.remotes if r.name == 'origin'), None)
            if origin_remote is None:
                return None
            return Project.org_from_remote_url(origin_remote.url)
        except InvalidGitRepositoryError:
            return None

    @staticmethod
    def config_dir() -> Path:
        ret = Path(CONFIG_DIR).expanduser()
        ret.mkdir(parents=True, exist_ok=True)
        return ret

    def prompt_org(self) -> str:
        default = self.detect_org()
        if default is None:
            choice = input('Organization: ')
            if not choice:
                raise InvalidOrgName()
            return choice
        return input(f'Organization ({default}): ') or default

    def tensorleap_dir(self) -> Path:
        return self.directory.joinpath(TENSORLEAP_DIR)

    @staticmethod
    def validate_project_name(name: str) -> None:
        if not name or re.match(VALID_PROJECT_REGEX, name) is None:
            raise InvalidProjectName()

    @staticmethod
    def validate_org_name(name: str) -> None:
        if not name or re.match(VALID_ORG_REGEX, name) is None \
                or '--' in name or len(name) < 3:
            raise InvalidOrgName()

    def is_initialized(self):
        return self.tensorleap_dir().is_dir()

    def init_project(
            self, framework: Framework, project: str, org: str, dataset: str, bucket: str,
            override_url: Optional[str] = None):
        if self.is_initialized():
            raise AlreadyInitialized()
        self.override_url = override_url
        self.project = project or self.prompt_project()
        Project.validate_project_name(self.project)
        self.org = org
        Project.validate_org_name(self.org)
        self.dataset = dataset
        self.framework = framework
        self.bucket = bucket
        self._generate_project_template()

    def config_file_path(self) -> Path:
        return self.tensorleap_dir().joinpath(CONFIG_FILENAME)

    def project_config(self) -> dict:
        with open(self.config_file_path(), 'r', encoding='utf-8') as config_file:
            _log.debug('reading stored config from %s', self.config_file_path())
            return toml.load(config_file)

    def _generate_project_template(self):
        assert self.project
        assert self.org
        assert not self.tensorleap_dir().is_dir()

        tgt = self.tensorleap_dir()
        copytree(Project.template_dir().joinpath(TENSORLEAP_DIR), tgt)
        txt = ''
        with open(self.config_file_path(), 'r', encoding='utf-8') as config_file:
            txt = config_file.read()
            txt = txt.replace('PROJ', self.project)
            txt = txt.replace('ORG', self.org)
            txt = txt.replace('FRAMEWORK', self.framework)
            txt = txt.replace('DATASET', self.dataset)
            txt = txt.replace('BUCKET', self.bucket)
            if self.override_url:
                if self.override_url[-1] == '/':
                    url_with_api_endpoint = f'{self.override_url}{API_ENDPOINT}'
                else:
                    url_with_api_endpoint = f'{self.override_url}/{API_ENDPOINT}'
                txt = txt.replace('API_ENDPOINT', url_with_api_endpoint)
            else:
                txt = txt.replace('API_ENDPOINT', '')

        with open(self.config_file_path(), 'w', encoding='utf-8') as config_file:
            config_file.write(txt)

    def _integration_file_py_path(self, file_name: str) -> Path:
        file_py = self.project_config()['integration'][file_name]
        return self.tensorleap_dir().joinpath(file_py)

    def model_py_path(self) -> Path:
        return self._integration_file_py_path('model')

    def dataset_py_path(self) -> Path:
        return self._integration_file_py_path('dataset')



import logging
import os
from pythonjsonlogger import jsonlogger

def configure_logging():
    log_level = int(os.environ.get('LEAP_LOG_LEVEL', str(logging.ERROR)))
    handler = logging.StreamHandler()
    formatter = jsonlogger.JsonFormatter()
    handler.setFormatter(formatter)
    logger = logging.getLogger()
    logger.setLevel(log_level)
    logger.addHandler(handler)



from pathlib import Path


def leap_save_model(target_file_path: Path):
    # Load your model
    # Save it to the path supplied as an arugment (has a .h5 suffix)

    print(f'Saving the model as {target_file_path}. Safe to delete this print.')



import importlib.util
import tempfile
import hashlib
import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import Tuple, Optional

import requests
from openapi_client.model.dataset_parse_request_params import DatasetParseRequestParams
from openapi_client.model.external_import_model_storage import ExternalImportModelStorage
from openapi_client.model.save_dataset_version_params import SaveDatasetVersionParams
from openapi_client.models import GetUploadSignedUrlParams, ImportNewModelParams, ImportModelType, \
    ExternalImportModelStorageResponse, GetCurrentProjectVersionParams

from leapcli.project import Project
from leapcli.login import Authenticator
from leapcli.exceptions import ModelNotFound, ModelEntryPointNotFound, ModelSaveFailure, \
    ModelNotSaved

_log = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class Push:
    def __init__(self, project: Project):
        self.project = project

        Authenticator.initialize(self.project)
        self._api = Authenticator.authenticated_api()

    @staticmethod
    def file_sha(path: Path) -> str:
        file_hash = hashlib.sha256()
        block_size = 2 ** 16
        with open(path, 'rb') as f:
            chunk = f.read(block_size)
            while len(chunk) > 0:
                file_hash.update(chunk)
                chunk = f.read(block_size)  # Read the next block from the file
            return file_hash.hexdigest()

    def get_import_url(self, filename: str) -> Tuple[str, str]:
        params = GetUploadSignedUrlParams(filename)
        import_url_response: ExternalImportModelStorage = self._api.get_upload_signed_url(params)
        return import_url_response.url, import_url_response.file_name

    @staticmethod
    def upload_file(url: str, path: Path) -> None:
        with open(path, 'rb') as f:
            requests.put(url, f, headers={"content-type": "application/octet-stream"})

    def current_project_version(self) -> str:
        proj_id = self.project.project_id(self._api)
        return self._api.get_current_project_version(GetCurrentProjectVersionParams(proj_id)).version_id

    def _create_import_new_model_params(
            self, filename: str, branch_name: Optional[str] = None,
            version: Optional[str] = None, model_name: Optional[str] = None) -> ImportNewModelParams:
        proj_id = self.project.project_id(self._api)
        if not version:
            version = self.current_project_version()
        if not model_name:
            model_name = f'CLI_{datetime.utcnow().strftime("%m/%d/%Y_%H:%M:%S")}'

        additional_arguments = {}
        if branch_name:
            additional_arguments['branch_name'] = branch_name

        return ImportNewModelParams(
            proj_id, filename, model_name, version, ImportModelType('H5_TF2'), **additional_arguments)

    def start_import_job(self, filename: str, branch_name: Optional[str] = None,
                         version: Optional[str] = None, model_name: Optional[str] = None) -> str:
        params = self._create_import_new_model_params(filename, branch_name, version, model_name)

        response: ExternalImportModelStorageResponse = self._api.import_model(params)
        return response.import_model_job_id

    def import_model(self, content_hash: str, path: Path, branch_name: Optional[str] = None,
                     version: Optional[str] = None, model_name: Optional[str] = None) -> str:
        url, file_name = self.get_import_url(content_hash)
        Push.upload_file(url, path)
        return self.start_import_job(file_name, branch_name, version, model_name)

    def _parse_dataset(self, dataset_py_as_string: str):
        dataset_id = self.project.dataset_id(self._api)
        params = DatasetParseRequestParams(dataset_id, is_public_bucket=False, script=dataset_py_as_string)
        self._api.parse_dataset(params)

    def _save_dataset(self, dataset_py_as_string: str):
        dataset_id = self.project.dataset_id(self._api)

        params = SaveDatasetVersionParams(dataset_id, bucket=self.project.detect_bucket(), script=dataset_py_as_string,
                                          save_as_new=False, is_public_bucket=False, is_valid=False)
        self._api.save_dataset_version(params)

    def save_and_parse_dataset(self):
        dataset_py_path = self.project.dataset_py_path()
        with open(str(dataset_py_path), "r") as f:
            dataset_py_as_string = f.read()
        self._save_dataset(dataset_py_as_string)

    @staticmethod
    def load_model_config_module(model_py_path: Path):
        if not model_py_path.is_file():
            raise ModelNotFound()

        spec = importlib.util.spec_from_file_location('tensorleap.model', model_py_path)
        model_module = importlib.util.module_from_spec(spec)

        sys.modules['tensorleap.model'] = model_module
        spec.loader.exec_module(model_module)
        return model_module

    # TODO: un-hardcode the .h5 suffix
    # Returns path to serialized model in cache dir and content content_hash of the file
    def serialize_model(self) -> Tuple[Path, str]:
        model_py_path = self.project.model_py_path()
        _log.debug('Looking for model integration file', extra=dict(path=model_py_path))

        if not model_py_path.is_file():
            raise ModelNotFound()

        _log.info('Loading user model configuration', extra=dict(path=model_py_path))
        model_module = Push.load_model_config_module(model_py_path)

        if not hasattr(model_module, 'leap_save_model'):
            raise ModelEntryPointNotFound()

        _, tmp_h5 = tempfile.mkstemp(suffix='.h5')
        tmp_h5 = Path(tmp_h5)

        _log.info('Invoking user leap_save_model', extra=dict(tgt_path=tmp_h5))
        try:
            model_module.leap_save_model(tmp_h5)
        except Exception as error:
            raise ModelSaveFailure() from error

        # Don't accumulate temp files with identical content
        # TODO: future enhancement: don't uploads to server if already uploaded before

        # File could exist but have 0 bytes because of mktemp
        if not tmp_h5.exists() or tmp_h5.stat().st_size == 0:
            raise ModelNotSaved()
        content_hash = Push.file_sha(tmp_h5)
        cache_path = self.project.cache_dir().joinpath(content_hash + '.h5')
        tmp_h5.rename(cache_path)

        return cache_path, content_hash

    def run(self, should_push_model=True, should_push_dataset=True,
            branch_name: Optional[str] = None, version: Optional[str] = None, model_name: Optional[str] = None):
        if should_push_model:
            path, content_hash = self.serialize_model()
            self.import_model(content_hash, path, branch_name, version, model_name)
        if should_push_dataset:
            self.save_and_parse_dataset()

        print('Push command successfully complete')



import logging
import sys
import traceback
from functools import lru_cache
from typing import Optional
from urllib3.exceptions import MaxRetryError


import luddite
import pkg_resources
from code_loader import DatasetLoader
from openapi_client.api.default_api import DefaultApi
from openapi_client.exceptions import NotFoundException, UnauthorizedException
from semver import VersionInfo

from leapcli.project import Project
from leapcli.login import Authenticator
from leapcli.exceptions import LoginFailed, DoctorCheckFailed
import leapcli.exceptions
from leapcli.push import Push

_log = logging.getLogger(__name__)

BANNER = '''
*******************************************************************
*                       Tensorleap CLI                            *
*                       --------------                            *
*                        health check                             *
*******************************************************************
'''

SCREEN_WIDTH = len(BANNER.split('\n')[1])


def format_kv(key: str, val: str) -> str:
    padding = SCREEN_WIDTH - len(key) - len(val) - 1
    assert padding > 0
    return f'{key}:{" " * padding}{val}'


# Prints key aligned to left and value to right.
def print_kv(key: str, val: str):
    print(format_kv(key, val))


def yesnobool(value: bool) -> str:
    return 'yes' if value else 'no'


class Doctor:
    def __init__(self, project: Project):
        self.project = project
        self._api: Optional[DefaultApi] = None

    def _get_api(self) -> DefaultApi:
        if self._api is None:
            self._api = Authenticator.authenticated_api()
        return self._api

    @staticmethod
    @lru_cache(1)
    def latest_cli_version() -> VersionInfo:
        return VersionInfo.parse(luddite.get_version_pypi("leapcli"))

    @staticmethod
    @lru_cache(1)
    def current_cli_version() -> VersionInfo:
        return VersionInfo.parse(pkg_resources.get_distribution("leapcli").version)

    @lru_cache(1)
    def update_available(self):
        return VersionInfo.compare(self.current_cli_version(),
                                   self.latest_cli_version()) < 0

    def check_cli(self):
        print_kv('Tensorleap CLI version', str(self.current_cli_version()))
        update = 'no'
        if self.update_available():
            update = f'yes ({str(self.latest_cli_version())})'
        print_kv('CLI update available', update)

    def check_project(self):
        print('\n* Inspecting project settings...\n')
        initialized = self.project.is_initialized()
        init = yesnobool(initialized)
        print_kv('Project initialized', init)
        if not initialized:
            print("\nRun `leap init` in your project's root directory.")
            raise DoctorCheckFailed()
        print_kv('Organization', self.project.detect_org())
        print_kv('API URL', self.project.api_url())
        print_kv('API Endpoint', self.project.detect_backend_url())
        print_kv('Project name', self.project.detect_project())
        print_kv('Framework', self.project.framework)
        print_kv('Dataset name', self.project.detect_dataset())
        print_kv('Bucket name', self.project.detect_bucket())

    def check_model(self) -> bool:
        path, _ = Push(self.project).serialize_model()
        return path.is_file()

    def check_dataset(self):
        dataset_py_path = self.project.dataset_py_path()
        with open(str(dataset_py_path), "r") as f:
            dataset_py_as_string = f.read()
        dataset_loader = DatasetLoader(dataset_py_as_string)
        dataset_parse_result = dataset_loader.check_dataset()
        if not dataset_parse_result.is_valid:
            print('\nFailed to parse dataset script.\n')
            error_message = dataset_parse_result.general_error
            if error_message is None:
                error_message = str([payload.display for payload in dataset_parse_result.payloads
                                     if not payload.is_passed])
            print(error_message)
            raise leapcli.exceptions.ParseDatasetFailed()
        print_kv('Dataset parse', 'ok')

    def check_codebase(self, should_check_model=True, should_check_dataset=True):
        print('\n* Inspecting model...\n')
        try:
            if should_check_model:
                if self.check_model():
                    print_kv('Model serialization', 'ok')
            if should_check_dataset:
                self.check_dataset()
            return
        except leapcli.exceptions.ModelNotFound:
            print_kv('Model serialization', 'error')
            expected_model_path = self.project.model_py_path()
            print(f'\nFailed to load model configuration.\n'
                  f'Path: {expected_model_path}')
        except leapcli.exceptions.ModelEntryPointNotFound:
            print_kv('Model serialization', 'error')
            expected_model_path = self.project.model_py_path()
            print(f'\nleap_save_model(path) function not found in model\n'
                  f'configuration file.\n'
                  f'\n'
                  f'Path: {expected_model_path}')
        except leapcli.exceptions.ModelNotSaved:
            print_kv('Model serialization', 'error')
            expected_model_path = self.project.model_py_path()
            print('\nleap_save_model(path) ran by did not save the model\n'
                  'to the given .h5 file.')
        except leapcli.exceptions.ModelSaveFailure as error:
            print_kv('Model serialization', 'error')
            print('\nleap_save_model(path) raised an exception.\n')
            print(type(error.inner_exception))
            traceback.print_exception(error.exc_type,
                                      error.inner_exception,
                                      error.traceback,
                                      file=sys.stdout)

        except Exception:  # pylint: disable=broad-except
            print_kv('Model serialization', 'error')
            print('Unknown error trying to save the model')
            traceback.print_exc()

        raise DoctorCheckFailed()

    def check_creds(self):
        print('\n* Inspecting credentials...')
        if not Authenticator.has_credentials():
            domain = self.project.org_domain()
            # TODO: use the real URL once this screen is implemented
            # in the web UI
            url = f'https://{domain}/profile/api-keys'
            print(f'\n'
                  f'Credentials not found.\n'
                  f'\n'
                  f'You can generate an API key at:\n'
                  f'  {url}\n')
            raise DoctorCheckFailed()
        try:
            Authenticator.initialize(self.project)

            print()
            print_kv('Logged in as', Authenticator.user.local.email)
            return True
        except (UnauthorizedException, NotFoundException, LoginFailed):
            print('\n'
                  'Login failed with the configured credentials.\n'
                  '\n'
                  'Try running `leap login` again.\n')
            raise DoctorCheckFailed()

    def check_project_exist(self):
        try:
            print('\n* Finding project...\n')
            project_id = self.project.project_id(self._get_api())
            print_kv("Project found. ID", project_id)
            return
        except AssertionError:
            print(f'Did not find project named {self.project.project}. Please create new project using the web UI or \n'
                  f'            run leap init and change to exist project.')
        raise DoctorCheckFailed()

    def check_dataset_exist(self):
        try:
            print('\n* Finding dataset...\n')
            dataset_id = self.project.dataset_id(self._get_api())
            print_kv("Dataset found. ID", dataset_id)
            return
        except AssertionError:
            print(f'Did not find dataset named {self.project.dataset}. Please create new dataset using the web UI or \n'
                  f'            run leap init and change to exist dataset.')
        raise DoctorCheckFailed()

    def run(self, should_check_model=True, should_check_dataset=True):
        print(BANNER)
        try:
            self.check_cli()
            self.check_project()
            self.check_creds()
            self.check_project_exist()
            if should_check_dataset:
                self.check_dataset_exist()
            self.check_codebase(should_check_model, should_check_dataset)
        except MaxRetryError:
            print('Failed to access the server. Please check your apiEndpoint in .tensorleap/config.toml')
            print('Check failed.')
        except DoctorCheckFailed:
            print('Check failed.')



"""Tensorleap CLI.

Usage:
  leap init (--tensorflow|--pytorch) (PROJECT) (ORG) (DATASET) (BUCKET) [--url=<URL>]
  leap login [API_ID] [API_KEY]
  leap check (--all|--exclude-model|--exclude-dataset)
  leap push (--all|--model|--dataset) [--branch-name=<BRANCH_NAME>]
            [--description=<DESCRIPTION>] [--model-name=<MODEL_NAME>]

Arguments:
  EXPERIMENT    Name of experiment.
  PROJECT       Project name (default: current directory name).
  ORG           Organization name (default: Git origin).

Options:
  -h --help     Show this screen.
  --version     Show version.
"""

import os
import sys
from typing import Callable, Dict
from pathlib import Path
from docopt import docopt
from openapi_client.exceptions import NotFoundException, UnauthorizedException
from leapcli.project import Project, Framework, VALID_PROJECT_EXPL, \
    VALID_ORG_EXPL, TENSORLEAP_DIR, CONFIG_FILENAME
from leapcli.exceptions import MalformedKeys, KeysMixedUp, AlreadyInitialized, \
    InvalidProjectName, InvalidOrgName
from leapcli.login import Authenticator
from leapcli.doctor import Doctor
from leapcli.log import configure_logging
from leapcli.push import Push


def __main__():
    # Add user repo path
    sys.path.insert(0, str(Path('.')))

    configure_logging()
    arguments = docopt(__doc__)
    if arguments['init']:
        init_command(arguments)
    if arguments['login']:
        login_command(arguments)
    if arguments['check']:
        check_command(arguments)
    if arguments['push']:
        push_command(arguments)


def eof_handler(func: Callable):
    def inner(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except EOFError:
            sys.exit(0)

    return inner


def complain(text: str):
    print(text, file=sys.stderr)
    sys.exit(1)


@eof_handler
def login_command(arguments: dict):
    project = Project(os.getcwd())
    if not project.is_initialized():
        complain('Tensorleap project not initialized.\n'
                 'Did you run `leap init`?')

    if arguments['API_ID'] is not None and arguments['API_KEY'] is None:
        complain(f'Either supply both API_ID and API_KEY or neither.\n'
                 f'{__doc__}')
    try:
        api_id = arguments.get('API_ID')
        api_key = arguments.get('API_KEY')
        if api_id is None:
            api_id = input('API ID: ')
            if api_key is None:
                api_key = input('API Key: ')

        Authenticator.initialize(project, api_id, api_key, should_write_credentials=True)
        print(f'Authenticated as {Authenticator.user.local.email}')

    except MalformedKeys:
        complain(f'Invalid API_ID or API_KEY.\n'
                 f'{__doc__}')
    except KeysMixedUp:
        complain(f'API_ID should come before API_KEY.\n'
                 f'{__doc__}')
    except NotFoundException:
        complain('❗️ Login failed. Check the API_ID provided.')
    except UnauthorizedException:
        complain('❗️ Login failed. Check the API_KEY provided.')


@eof_handler
def check_command(arguments: Dict):
    should_check_model = not arguments['--exclude-model']
    should_check_dataset = not arguments['--exclude-dataset']
    project = Project(os.getcwd())
    Doctor(project).run(should_check_model, should_check_dataset)


@eof_handler
def push_command(arguments):
    should_push_model = arguments['--all'] or arguments['--model']
    should_push_dataset = arguments['--all'] or arguments['--dataset']

    project = Project(os.getcwd())
    Push(project).run(should_push_model, should_push_dataset,
                      arguments.get('--branch-name'),
                      arguments.get('--description'),
                      arguments.get('--model-name'))


@eof_handler
def init_command(arguments: Dict):
    framework = Framework.TENSORFLOW if arguments['--tensorflow'] else Framework.PYTORCH
    try:
        initializer = Project(os.getcwd())
        initializer.init_project(framework,
                                 arguments.get('PROJECT'),
                                 arguments.get('ORG'),
                                 arguments.get('DATASET'),
                                 arguments.get('BUCKET'),
                                 arguments.get('--url'))
        print(f'Tensorleap project initialized in {TENSORLEAP_DIR}')
    except AlreadyInitialized:
        expected_conf = Path(TENSORLEAP_DIR).joinpath(CONFIG_FILENAME)
        complain(f'Tensorleap project already initialized.\n'
                 f'See {expected_conf}')
    except InvalidProjectName:
        complain(f'Invalid project name. Rules:\n{VALID_PROJECT_EXPL}')
    except InvalidOrgName:
        complain(f'Invalid organization name. Rules:\n{VALID_ORG_EXPL}')


if __name__ == '__main__':
    __main__()



import sys


class AlreadyInitialized(Exception):
    pass


class InvalidProjectName(Exception):
    pass


class ProjectNotInitialized(Exception):
    pass


class CredentialsNotFound(Exception):
    pass


class InvalidOrgName(Exception):
    pass


class MalformedKeys(Exception):
    pass


class KeysMixedUp(Exception):
    pass


class LoginFailed(Exception):
    pass


class ModelNotFound(Exception):
    pass


class ParseDatasetFailed(Exception):
    pass


class ModelEntryPointNotFound(Exception):
    pass


class ModelSaveFailure(Exception):
    def __init__(self):
        super().__init__()
        self.exc_type, self.inner_exception, self.traceback = sys.exc_info()


class DoctorCheckFailed(Exception):
    def __init__(self):
        super().__init__()
        self.exc_type, self.inner_exception, self.traceback = sys.exc_info()


class ModelNotSaved(Exception):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['leapcli']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.18,<4.0.0',
 'code-loader==0.1.15',
 'docopt>=0.6.2,<0.7.0',
 'luddite>=1.0.2,<2.0.0',
 'pytest-mock>=3.6.1,<4.0.0',
 'python-json-logger>=2.0.2,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'semver>=2.13.0,<3.0.0',
 'tensorleap-openapi-client==1.1.9',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['leap = cli:__main__']}

setup_kwargs = {
    'name': 'leapcli',
    'version': '0.2.8',
    'description': 'Tensorleap CLI',
    'long_description': None,
    'author': 'Assaf Lavie',
    'author_email': 'assaf.lavie@tensorleap.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tensorleap/cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<3.7',
}


setup(**setup_kwargs)



import os
from typing import Union, List
import math
import numpy as np

from code_loader.contract.enums import DatasetInputType, DatasetOutputType, DatasetMetadataType
from code_loader.contract.datasetclasses import SubsetResponse
from code_loader import dataset_binder
from google.cloud import storage
from google.cloud.storage import Bucket
from google.oauth2 import service_account

crop_size = 128


def _connect_to_gcs() -> Bucket:
    auth_secret = os.environ['AUTH_SECRET']
    if type(auth_secret) is dict:
        # getting credentials from dictionary account info
        credentials = service_account.Credentials.from_service_account_info(auth_secret)
    else:
        # getting credentials from path
        credentials = service_account.Credentials.from_service_account_file(auth_secret)
    project = credentials.project_id

    gcs_client = storage.Client(project=project, credentials=credentials)

    bucket_name = 'YOUR_BUCKET_NAME'

    return gcs_client.bucket(bucket_name)


def subset_func() -> List[SubsetResponse]:
    bucket = _connect_to_gcs()

    base_path = 'ISBICellSegmentation/'
    train_path = f'{base_path}train-volume2.npy'

    blob = bucket.blob(train_path)
    blob.download_to_filename(train_path)
    scans = np.load(train_path) / 255

    train_path = f'{base_path}train-labels2.npy'
    blob.download_to_filename(train_path)
    masks = np.load(train_path) / 255
    image_size = scans.shape[-1]

    val_split = int(len(scans) * 0.8)
    train_scans = scans[:val_split]
    val_scans = scans[val_split:]
    train_masks = masks[:val_split]
    val_masks = masks[val_split:]

    crop_per_image = int(image_size / crop_size) ** 2
    train_length = crop_per_image * len(train_scans)
    val_length = crop_per_image * len(val_scans)
    crop_per_image = int(image_size / crop_size) ** 2

    train = SubsetResponse(length=train_length, data={'images': train_scans, 'masks': train_masks,
                                                      'shape': image_size, 'crop_per_image': crop_per_image})
    val = SubsetResponse(length=val_length, data={'images': val_scans, 'masks': val_masks, 'shape': image_size,
                                                  'crop_per_image': crop_per_image})
    response = [train, val]
    return response


def input_encoder_1(idx: int, subset: SubsetResponse) -> np.ndarray:
    crop_per_image = subset.data['crop_per_image']
    image_idx = int(idx / crop_per_image)
    image = subset.data['images'][image_idx]
    img_slice = idx - (image_idx * crop_per_image)
    crop = np.split(image, crop_per_image)[img_slice]

    axis_slice = int(math.sqrt(crop_per_image))

    hi = int(img_slice / axis_slice)
    vi = int(img_slice % axis_slice)
    crop = np.hsplit(np.vsplit(image, axis_slice)[hi], axis_slice)[vi]

    return crop[..., np.newaxis]


def output_encoder_1(idx: int, subset: Union[SubsetResponse, list]) -> np.ndarray:
    crop_per_image = subset.data['crop_per_image']
    image_idx = int(idx / crop_per_image)
    image = subset.data['masks'][image_idx]
    img_slice = idx - (image_idx * crop_per_image)
    crop = np.split(image, crop_per_image)[img_slice]

    axis_slice = int(math.sqrt(crop_per_image))

    hi = int(img_slice / axis_slice)
    vi = int(img_slice % axis_slice)
    crop = np.hsplit(np.vsplit(image, axis_slice)[hi], axis_slice)[vi]

    return crop[..., np.newaxis]


def metadata_encoder_1(idx: int, subset: Union[SubsetResponse, list]) -> np.ndarray:
    crop_per_image = subset.data['crop_per_image']
    image_idx = int(idx / crop_per_image)
    image = subset.data['masks'][image_idx]
    img_slice = idx - (image_idx * crop_per_image)
    crop = np.split(image, crop_per_image)[img_slice]

    axis_slice = int(math.sqrt(crop_per_image))

    hi = int(img_slice / axis_slice)
    vi = int(img_slice % axis_slice)
    crop = np.hsplit(np.vsplit(image, axis_slice)[hi], axis_slice)[vi]
    return float(np.mean(crop))


def metadata_encoder_2(idx: int, subset: Union[SubsetResponse, list]) -> np.ndarray:
    crop_per_image = subset.data['crop_per_image']
    image_idx = int(idx / crop_per_image)
    image = subset.data['masks'][image_idx]
    img_slice = idx - (image_idx * crop_per_image)
    crop = np.split(image, crop_per_image)[img_slice]

    axis_slice = int(math.sqrt(crop_per_image))

    hi = int(img_slice / axis_slice)
    vi = int(img_slice % axis_slice)
    crop = np.hsplit(np.vsplit(image, axis_slice)[hi], axis_slice)[vi]
    return float(np.std(crop))


dataset_binder.set_subset(ration=1, function=subset_func, name='MicroscopeCellImages')

dataset_binder.set_input(function=input_encoder_1, subset='MicroscopeCellImages', input_type=DatasetInputType.Image,
                         name='Crops')

dataset_binder.set_ground_truth(function=output_encoder_1, subset='MicroscopeCellImages',
                                ground_truth_type=DatasetOutputType.Mask,
                                name='segments', labels=['Background', 'Cell'], masked_input='Crops')

dataset_binder.set_metadata(function=metadata_encoder_1, subset='MicroscopeCellImages',
                            metadata_type=DatasetMetadataType.float,
                            name='label_mean')
dataset_binder.set_metadata(function=metadata_encoder_2, subset='MicroscopeCellImages',
                            metadata_type=DatasetMetadataType.float,
                            name='laebl_std')
